implement Emuinit;

include "sys.m";
	sys: Sys;

include "draw.m";
include "sh.m";
include "arg.m";
	arg: Arg;

Emuinit: module
{
	init: fn();
};

init()
{
	sys = load Sys Sys->PATH;
	sys->bind("#e", "/env", sys->MREPL|sys->MCREATE);	# if #e not configured, that's fine
	args := getargs();
	cmd := Command->PATH;
	arg = load Arg Arg->PATH;
	if (arg == nil)
		sys->fprint(sys->fildes(2), "emuinit: unable to load Arg\n");
	else {
		arg->init(args);
		while((c := arg->opt()) != 0)
			case c {
			'g' or 'c' or 'C' or 'm' or 'p' or 'f' or 'r' or 'd' => arg->arg();
	                  }
		args = arg->argv();
	}
	if (args != nil)
		cmd = hd args;
	sh: Command;
	if (cmd[0] == '/')
		sh = load Command cmd;
	else {
		sh = load Command "/dis/"+cmd;
		if (sh == nil)
			sh = load Command "/"+cmd;
	}
	if (cmd == Command->PATH) {	# add startup option
		if (args == nil)
			args = cmd :: "-l" :: nil;
		else
			args = cmd :: "-l" :: tl args;
	}
	if (sh == nil)
		sys->fprint(sys->fildes(2), "emuinit: unable to load %s: %r\n", cmd);
	else
		sh->init(nil, args);
}

getargs(): list of string
{
	fd := sys->open("#e/emuargs", Sys->OREAD);
	if (fd == nil)
		return nil;
	maxargs := 1024;
	(ok, d) := sys->fstat(fd);
	if(ok >= 0 && int d.length > maxargs)
		maxargs = int d.length;
	buf := array[maxargs] of byte;
	n := sys->read(fd, buf, len buf);
	if (n <= 0)
		return nil;
	return unquoted(string buf[0:n]);
}

#
# this version was faster when last tested
# against a more straightforward version
#
# we duplicate this from String so that
# emuinit will run with just Sys
#
unquoted(args: string): list of string
{
	OUTWORD, INWORD, INQUOTE, SEENQUOTE: con iota;
	argv: list of string;
	word: string;
	mark := -1;
	state := OUTWORD;
	for (i := len args - 1; i >= 0; i--) {
		c := args[i];
		# if statements are much faster than case statements in compiled mode.
		if (state == INWORD) {
			if (c == '\'') {
				argv = args[i + 1:mark] :: argv;
				mark = i;
				word = nil;
				state = INQUOTE;
			} else if (c == ' ' || c == '\t' || c == '\n') {
				argv = args[i + 1:mark] :: argv;
				state = OUTWORD;
			}
		} else if (state == INQUOTE) {
			if (c == '\'') {
				word = args[i + 1:mark] + word;
				state = SEENQUOTE;
			}
		} else if (state == SEENQUOTE) {
			if (c == ' ' || c == '\t' || c == '\n') {
				argv = word :: argv;
				state = OUTWORD;
			} else if (c == '\'') {
				mark = i + 1;
				state = INQUOTE;
			} else {
				argv = word :: argv;
				mark = i + 1;
				state = INWORD;
			}
		} else {		# OUTWORD
			if (c == '\'') {
				word = nil;
				mark = i;
				state = INQUOTE;
			} else if (c != ' ' && c != '\t'&& c != '\n') {
				mark = i + 1;
				state = INWORD;
			}
		}
	}
	if (state == INWORD)
		argv = args[0:mark] :: argv;
	else if (state == INQUOTE)
		argv = nil;		# unbalanced; nil is better than confused quoted text.
	else if (state == SEENQUOTE)
		argv = word :: argv;
	return argv;
}
