implement Fsmodule;
include "sys.m";
	sys: Sys;
include "draw.m";
include "sh.m";
include "fslib.m";
	fslib: Fslib;
	Report, Value, type2s: import fslib;
	Fschan, Fsdata, Entrychan, Entry,
	Gatechan, Gatequery, Nilentry, Option,
	Next, Down, Skip, Quit: import Fslib;

types(): string
{
	return "xxxx*-1";
}

init()
{
	sys = load Sys Sys->PATH;
	fslib = load Fslib Fslib->PATH;
	if(fslib == nil){
		sys->fprint(sys->fildes(2), "fs: cannot load %s: %r\n", Fslib->PATH);
		raise "fail:bad module";
	}
}

run(nil: ref Draw->Context, nil: ref Report,
			opts: list of Option, args: list of ref Value): ref Value
{
	dst := chan of (Fsdata, chan of int);
	spawn mergeproc((hd args).x().i, (hd tl args).x().i, dst, opts == nil);
	for(args = tl tl args; args != nil; args = tl args){
		dst1 := chan of (Fsdata, chan of int);
		spawn mergeproc(dst, (hd args).x().i, dst1, opts == nil);
		dst = dst1;
	}
	return ref Value.X(dst);
}

# merge two trees; assume directories are alphabetically sorted.
mergeproc(c0, c1, dst: Fschan, recurse: int)
{
	d0, d1: ref Sys->Dir;
	reply: chan of int;

	myreply := chan of int;

	# use the root directory from the first stream, and ignore the other.
	((d0, nil), reply) = <-c0;
	dst <-= ((d0, nil), myreply);
	if((reply <-= <-myreply) != Down){
		(<-c1).t1 <-= Quit;
		exit;
	}

	((d1, nil), reply) = <-c1;
	reply <-= Down;

	mergedir(c0, c1, dst, recurse);
}

mergedir(c0, c1, dst: Fschan, recurse: int)
{
	myreply := chan of int;
	reply0, reply1: chan of int;
	d0, d1: ref Sys->Dir;
	for(;;){
		if(d0 == nil){
			((d0, nil), reply0) = <-c0;
		}
		if(d1 == nil){
			((d1, nil), reply1) = <-c1;
		}
		if(d0 == nil || d1 == nil)
			break;
		wc0, wc1:	Fschan;
		wd0, wd1: ref Sys->Dir;
		wreply0, wreply1: chan of int;
		if(d0.name == d1.name){
			if((d0.mode & d1.mode & Sys->DMDIR) != 0 && recurse){
				dst <-= ((d0, nil), myreply);
				r := <-myreply;
				reply0 <-= r;
				reply1 <-= r;
				d0 = d1 = nil;
				case r {
				Quit =>
					exit;
				Skip =>
					return;
				Down =>
					mergedir(c0, c1, dst, 1);
				}
				continue;
			}
			reply1 <-= Next;			# ignore c1's contribution
			(wd0, wc0, wreply0) = (d0, c0, reply0);
			(wd1, wc1, wreply1) = (nil, c1, nil);
			d0 = d1 = nil;
		}else if(d0.name < d1.name){
			(wd0, wc0, wreply0) = (d0, c0, reply0);
			(wd1, wc1, wreply1) = (d1, c1, reply1);
			d0 = nil;
		}else{
			(wd0, wc0, wreply0) = (d1, c1, reply1);
			(wd1, wc1, wreply1) = (d0, c0, reply0);
			d1 = nil;
		}
		dst <-= ((wd0, nil), myreply);
		r := <-myreply;
		wreply0 <-= r;
		if(r == Down)
			r = fslib->copy(wc0, dst);
		case r {
		Quit or
		Skip =>
			if(wd1 == nil)
				wreply1 = (<-wc1).t1;
			wreply1 <-= r;
			if(r == Quit)
				exit;
			return;
		}
	}
	# one directory has finished; copy the rest of the other.
	if(d0 == nil){
		d0 = d1;
		c0 = c1;
		(reply0, reply1) = (reply1, reply0);
	}
	reply1 <-= Next;
	if(d0 != nil){
		dst <-= ((d0, nil), myreply);
		r := <-myreply;
		if((reply0 <-= r) == Down)
			r = fslib->copy(c0, dst);
		if(r == Quit)
			exit;
		if(fslib->copy(c0, dst) == Quit)
			exit;
	}else{
		reply0 <-= Next;
		dst <-= ((nil, nil), myreply);
		if(<-myreply == Quit)
			exit;
	}
}
