implement Os;

include "sys.m";
	sys: Sys;

include "draw.m";

include "string.m";
	str: String;

include "arg.m";

Os: module
{
	init:	fn(nil: ref Draw->Context, nil: list of string);
};

init(nil: ref Draw->Context, args: list of string)
{
	sys = load Sys Sys->PATH;
	str = load String String->PATH;
	if(str == nil)
		fail(sys->sprint("cannot load %s: %r", String->PATH));
	arg := load Arg Arg->PATH;
	if(arg == nil)
		fail(sys->sprint("cannot load %s: %r", Arg->PATH));

	arg->init(args);
	arg->setusage("os [-d dir] [-n] command [arg...]");

	nice := 0;
	workdir := "";
	while((opt := arg->opt()) != 0) {
		case opt {
		'd' =>
			workdir = arg->earg();
		'n' =>
			nice = 1;
		* =>
			arg->usage();
		}
	}
	args = arg->argv();
	if (args == nil)
		arg->usage();
	arg = nil;

	sys->pctl(Sys->FORKNS, nil);
	sys->bind("#p", "/prog", Sys->MREPL);		# don't worry if it fails
	if(sys->stat("/cmd/clone").t0 != 0) {
		if(sys->bind("#C", "/", Sys->MBEFORE) < 0)
			fail(sys->sprint("bind #C /: %r"));
	}

	cmd := "exec";
	for(; args != nil; args = tl args)
		cmd += sys->sprint(" %q", hd args);
	cfd := sys->open("/cmd/clone", sys->ORDWR);
	if(cfd == nil)
		fail(sys->sprint("open /cmd/clone: %r"));
	
	buf := array[32] of byte;
	n := sys->read(cfd, buf, len buf);
	if(n <= 0)
		fail(sys->sprint("read /cmd/clone: %r"));

	dir := "/cmd/"+string buf[0:n];

	wfd := sys->open(dir+"/wait", Sys->OREAD);
	if(nice && sys->fprint(cfd, "nice") < 0)
		sys->fprint(sys->fildes(2), "os: can't set nice priority: %r\n");
	if(workdir != nil && sys->fprint(cfd, "dir %s", workdir) < 0)
		fail(sys->sprint("host cd %s: %r", workdir));
	if(sys->fprint(cfd, "%s", cmd) < 0)
		fail(sys->sprint("exec: %r"));
	tocmd := sys->open(dir+"/data", sys->OWRITE);
	if(tocmd == nil)
		fail(sys->sprint("can't open %s/data for writing: %r", dir));
	fromcmd := sys->open(dir+"/data", sys->OREAD);
	if(fromcmd == nil)
		fail(sys->sprint("can't open %s/data for reading: %r", dir));
	cfd = nil;

	sync := chan[1] of int;
	spawn copy(sync, sys->fildes(0), tocmd);
	pid := <-sync;
	tocmd = nil;
	copy(sync, fromcmd, sys->fildes(1));
	kill(pid);

	if(wfd != nil){
		status := array[1024] of byte;
		n = sys->read(wfd, status, len status);
		if(n < 0)
			fail(sys->sprint("wait error: %r"));
		s := string status[0:n];
		if(s != nil){
			# pid user sys real status
			flds := str->unquoted(s);
			if(len flds < 5)
				fail(sys->sprint("wait error: odd status: %q", s));
			s = hd tl tl tl tl flds;
			if(0)
				sys->fprint(sys->fildes(2), "WAIT: %q\n", s);
			if(s != nil)
				raise "fail:host: "+s;
		}
	}
}

copy(sync: chan of int, f, t: ref Sys->FD)
{
	sync <-= sys->pctl(0, nil);
	buf := array[8192] of byte;
	for(;;) {
		r := sys->read(f, buf, len buf);
		if(r <= 0)
			break;
		w := sys->write(t, buf, r);
		if(w != r)
			break;
	}
}

kill(pid: int)
{
	fd := sys->open("/prog/"+string pid+"/ctl", sys->OWRITE);
	sys->fprint(fd, "kill");
}

fail(msg: string)
{
	sys->fprint(sys->fildes(2), "os: %s\n", msg);
	raise "fail:"+msg;
}
