implement Diskblocks;

#
# adapted from Acme's disk.b
#

include "sys.m";
	sys: Sys;

include "diskblocks.m";

init()
{
	sys = load Sys Sys->PATH;
}

tempfile(): ref Sys->FD
{
	buf := sys->sprint("/tmp/X%d.%.4sblks", sys->pctl(0, nil), "inferno");	# TO DO
	for(i:='A'; i<='Z'; i++){
		buf[5] = i;
		(ok, nil) := sys->stat(buf);
		if(ok == 0)
			continue;
		fd := sys->create(buf, Sys->ORDWR|Sys->ORCLOSE, 8r600);
		if(fd != nil)
			return fd;
	}
	return nil;
}

Disk.init(fd: ref Sys->FD, maxblock: int, gran: int): ref Disk
{
	d := ref Disk;
	if(gran == 0 || maxblock%gran != 0)
		return nil;
	d.maxblock = maxblock;
	d.gran = gran;
	d.free = array[maxblock/gran+1] of list of ref Block;
	d.addr = big 0;
	d.fd = fd;
	return d;
}

ntosize(d: ref Disk, n: int): (int, int)
{
	if (n > d.maxblock)
		return (-1, -1);
	size := n;
	if((size % d.gran) != 0)
		size += d.gran - size%d.gran;
	# last bucket holds blocks of exactly d.maxblock
	return (size, size/d.gran);
}

Disk.new(d: self ref Disk, n: int): ref Block
{
	(size, i) := ntosize(d, n);
	b: ref Block;
	if(d.free[i] != nil){
		b = hd d.free[i];
		d.free[i] = tl d.free[i];
	}else{
		b = ref Block(d.addr, 0);
		d.addr += big size;
	}
	b.n = n;
	return b;
}

Disk.release(d: self ref Disk, b: ref Block)
{
	(sz, i) := ntosize(d, b.n);
	d.free[i] = b :: d.free[i];
}

Disk.write(d: self ref Disk, b: ref Block, a: array of byte, n: int): (ref Block, string)
{
	if(b != nil){
		(size, nil) := ntosize(d, b.n);
		(nsize, nil) := ntosize(d, n);
		if(size != nsize){
			d.release(b);
			b = d.new(n);
		}
	}else
		b = d.new(n);
	if(b == nil)
		return (nil, "illegal diskblock allocation");
	if(sys->seek(d.fd, b.addr, 0) < big 0)
		return (b, sys->sprint("seek error: %r"));
	if(sys->write(d.fd, a, n) != n)
		return (b, sys->sprint("write error: %r"));
	b.n = n;
	return (b, nil);
}

Disk.read(d: self ref Disk, b: ref Block, a: array of byte, n: int): string
{
	if(n > b.n || n > d.maxblock)
		return "internal error: request bigger than Block";
	if(sys->seek(d.fd, big b.addr, 0) < big 0)
		return sys->sprint("seek error: %r");
	if(sys->read(d.fd, a, n) != n)
		return sys->sprint("read error: %r");
	return nil;
}
