implement UsbDriver;

include "sys.m";
	sys: Sys;
include "usb.m";
	usb: Usb;

readerpid: int;
setupfd, ctlfd: ref Sys->FD;
infd, outfd: ref Sys->FD;
globaltag: int;

lun: int;
fileio: ref Sys->FileIO;
blocksize: int;

kill(pid: int): int
{
	fd := sys->open("/prog/"+string pid+"/ctl", Sys->OWRITE);
	if (fd == nil)
		return -1;
	if (sys->write(fd, array of byte "kill", 4) != 4)
		return -1;
	return 0;
}

reader(pidc: chan of int)
{
	pid := sys->pctl(0, nil);
	pidc <-= pid;
	for (;;) {
		(offset, count, fid, rc) := <- fileio.read;
		if (rc != nil) {
sys->print("read offset %d length %d\n", offset, count);
			if (offset % blocksize) {
				rc <- = (nil, "unaligned");
				continue;
			}
			if (count % blocksize) {
				rc <- = (nil, "unaligned");
				continue;
			}
			offset /= blocksize;
			count /= blocksize;
sys->print("read block %d count %d\n", offset, count);
			buf := array [count * blocksize] of byte;
			if (scsiread10(lun, offset, count, buf) < 0) {
				scsirequestsense(lun);
				rc <- = (nil, "read error");
				continue;
			}
			rc <- = (buf, nil);
		}
	}
	readerpid = -1;
}

bulkonlymassstoragereset(): int
{
	if (usb->setup(setupfd, Usb->RH2D | Usb->Rclass | Usb->Rinterface, 255, 0, 0, nil, nil) < 0) {
		sys->print("usbmass: bulkonlystoragereset failed\n");
		return -1;
	}
	return 0;
}
	
bulkonlygetmaxlun(): int
{
	buf := array[1] of byte;
	if (usb->setup(setupfd, Usb->RD2H | Usb->Rclass | Usb->Rinterface, 254, 0, 0, nil, buf) < 0) {
		sys->print("usbmass: bulkonlygetmaxlun failed\n");
		return -1;
	}
	return int buf[0];
}

bulkonlysendcbw(tag: int, dtl: int, outdir: int, lun: int, cmd: array of byte): int
{
	cbw := array [31] of byte;
	cbw[0] = byte 16r55;
	cbw[1] = byte 16r53;
	cbw[2] = byte 16r42;
	cbw[3] = byte 16r43;
	usb->put4(cbw[4:], tag);
	usb->put4(cbw[8:], dtl);
	if (outdir)
		cbw[12] = byte 0;
	else
		cbw[12] = byte 16r80;
	cbw[13] = byte lun;
	cbw[14] = byte len cmd;
	cbw[15:] = cmd;
	rv := sys->write(outfd, cbw, len cbw);
	if (rv < 0) {
		sys->print("bulkonlysendcbw: failed: %r\n");
		return -1;
	}
	if (rv != len cbw) {
		sys->print("bulkonlysendcbw: truncated send\n");
		return -1;
	}
	return 0;
}

bulkonlyrecvcsw(tag: int): (int, int)
{
	sys->print("bulkonlyrecvcsw\n");
	buf := array [13] of byte;
	if (sys->read(infd, buf, len buf) != len buf) {
		sys->print("bulkonlyrecvcsw: read failed: %r\n");
		return (-1, -1);
	}
	if (usb->get4(buf) != 16r53425355) {
		sys->print("bulkonlyrecvcsw: signature wrong\n");
		return (-1, -1);
	}
	recvtag := usb->get4(buf[4:]);
	residue := usb->get4(buf[8:]);
	status := int buf[12];
	if (recvtag != tag) {
		sys->print("bulkonlyrecvcsw: tag does not match: sent %d recved %d\n", tag, recvtag);
		return (-1, -1);
	}
	sys->print("bulkonlyrecvcsw: residue %d status %d\n", residue, status);
	return (residue, status);
}

unstallread()
{
	sys->print("unstalling bulk-in\n");
#	sys->sleep(2000);
	if (usb->setclear_feature(setupfd, Usb->Rendpt, 0, 2, 0) < 0) {
		sys->print("unstallread: clear_feature() failed: %r\n");
		return;
	}
#	sys->sleep(2000);
#	sys->print("unstalling bulk-in phase 2\n");
	if (sys->fprint(ctlfd, "unstall 2") < 0)
		sys->print("unstallread: unstall failed: %r\n");
#	sys->print("unstalling bulk-in done\n");
}

warnfprint(fd: ref Sys->FD, s: string)
{
	if (sys->fprint(fd, "%s", s) != len s) {
		sys->print("warning: writing %s failed: %r\n", s);
	}
}

bulkreadcommand(lun: int, cmd: array of byte, buf: array of byte, dump: int): int
{
	globaltag = globaltag + 1;
	if (bulkonlysendcbw(globaltag, len buf, 0, lun, cmd) < 0)
		return -1;
	got := 0;
	if (buf != nil) {
		while (got < len buf) {
			rv := sys->read(infd, buf[got:], len buf - got);
			if (rv < 0) {
				sys->print("bulkreadcommand: read failed: %r\n");
				break;
			}
			sys->print("read %d\n", rv);
			got += rv;
			break;
		}
		if (dump) {
			for (i := 0; i < got; i++)
				sys->print("%.2ux", int buf[i]);
			sys->print("\n");
		}
		if (got == 0)
			unstallread();
	}
	(residue, status) := bulkonlyrecvcsw(globaltag);
	if (residue < 0) {
		unstallread();
		(residue, status) = bulkonlyrecvcsw(globaltag);
		if (residue < 0)
			return -1;
	}
	if (status != 0)
		return -1;
	return got;
}

scsiinquiry(lun: int): int
{
	buf := array [255] of byte;
	cmd := array [6] of byte;
	cmd[0] = byte 16r12;
	cmd[1] = 	byte (lun << 5);
	cmd[2] = byte 0;
	cmd[3] = byte 0;
	cmd[4] = byte len buf;
	cmd[5]  = byte 0;
	got := bulkreadcommand(lun, cmd, buf, 0);
	if (got < 0)
		return -1;
	if (got < 36) {
		sys->print("scsiinquiry: too little data\n");
		return -1;
	}
	t := int buf[0] & 16r1f;
	sys->print("scsiinquiry: type %d/%s\n",
	    t, string buf[8:35]);
	if (t != 0)
		return -1;
	return 0;
}

scsireadcapacity(lun: int): int
{
#	warnfprint(ctlfd, "debug 0 1");
#	warnfprint(ctlfd, "debug 1 1");
#	warnfprint(ctlfd, "debug 2 1");
	buf := array [8] of byte;
	cmd := array [10] of byte;
	cmd[0] = byte 16r25;
	cmd[1] = 	byte (lun << 5);
	cmd[2] = byte 0;
	cmd[3] = byte 0;
	cmd[4] = byte 0;
	cmd[5]  = byte 0;
	cmd[6]  = byte 0;
	cmd[7]  = byte 0;
	cmd[8]  = byte 0;
	cmd[9] = byte 0;
	got := bulkreadcommand(lun, cmd, buf, 0);
	if (got < 0)
		return -1;
	if (got != len buf) {
		sys->print("scsireadcapacity: returned data not right size\n");
		return -1;
	}
	blocksize = usb->bigget4(buf[4:]);
	sys->print("block size %d block %d\n", blocksize, usb->bigget4(buf[0:]));
	return 0;
}

scsirequestsense(lun: int): int
{
#	warnfprint(ctlfd, "debug 0 1");
#	warnfprint(ctlfd, "debug 1 1");
#	warnfprint(ctlfd, "debug 2 1");
	buf := array [255] of byte;
	cmd := array [6] of byte;
	cmd[0] = byte 16r03;
	cmd[1] = 	byte (lun << 5);
	cmd[2] = byte 0;
	cmd[3] = byte 0;
	cmd[4] = byte len buf;
	cmd[5]  = byte 0;
	got := bulkreadcommand(lun, cmd, buf, 1);
	if (got < 0)
		return -1;
	return 0;
}

scsiread10(lun: int, offset, count: int, buf: array of byte): int
{
#	warnfprint(ctlfd, "debug 0 1");
#	warnfprint(ctlfd, "debug 1 1");
#	warnfprint(ctlfd, "debug 2 1");
	cmd := array [10] of byte;
	cmd[0] = byte 16r28;
	cmd[1] = byte (lun << 5);
	usb->bigput4(cmd[2:], offset);
	cmd[6] = byte 0;
	usb->bigput2(cmd[7:], count);
	cmd[9] = byte 0;
	got := bulkreadcommand(lun, cmd, buf, 0);
	if (got < 0)
		return -1;
	return 0;
}

scsistartunit(lun: int, start: int): int
{
#	warnfprint(ctlfd, "debug 0 1");
#	warnfprint(ctlfd, "debug 1 1");
#	warnfprint(ctlfd, "debug 2 1");
	cmd := array [6] of byte;
	cmd[0] = byte 16r1b;
	cmd[1] = byte (lun << 5);
	cmd[2] = byte 0;
	cmd[3] = byte 0;
	cmd[4] = byte (start & 1);
	cmd[5]  = byte 0;
	got := bulkreadcommand(lun, cmd, nil, 0);
	if (got < 0)
		return -1;
	return 0;
}

init(usbmod: Usb, psetupfd, pctlfd: ref Sys->FD,
	nil: ref Usb->Device,
	conf: array of ref Usb->Configuration, path: string): int
{
	rv: int;

	usb = usbmod;
	setupfd = psetupfd;
	ctlfd = pctlfd;

	sys = load Sys Sys->PATH;
	rv = usb->set_configuration(setupfd, conf[0].id);
	if (rv < 0)
		return rv;
	rv = bulkonlymassstoragereset();
	if (rv < 0)
		return rv;
	maxlun := bulkonlygetmaxlun();
	if (maxlun < 0)
		return maxlun;
	lun = 0;
	sys->print("maxlun %d\n", maxlun);
	ep := (hd conf[0].iface[0].altiface).ep[0];
	sys->print("maxpkt %d interval %d\n", ep.maxpkt, ep.interval);
	rv = sys->fprint(ctlfd, "ep 1 %d w bulk 1", ep.maxpkt);
	if (rv < 0)
		return rv;
	ep = (hd conf[0].iface[0].altiface).ep[1];
	sys->print("maxpkt %d interval %d\n", ep.maxpkt, ep.interval);
	rv = sys->fprint(ctlfd, "ep 2 %d r bulk 32", ep.maxpkt);
	if (rv < 0)
		return rv;
	outfd = sys->open(path + "ep1data", Sys->OWRITE);
	if (outfd == nil)
		return -1;
	infd = sys->open(path + "ep2data", Sys->OREAD);
	if (infd == nil)
		return -1;
	if (scsiinquiry(0) < 0)
		return -1;
	if (scsireadcapacity(0) < 0) {
		scsirequestsense(0);
		if (scsireadcapacity(0) < 0)
			return -1;
	}
	if (scsistartunit(lun, 1) < 0) {
		return -1;
	}
	fileio = sys->file2chan("/chan", "usbdisk");
	if (fileio == nil) {
		sys->print("file2chan failed: %r\n");
		return -1;
	}
#	warnfprint(ctlfd, "debug 0 1");
#	warnfprint(ctlfd, "debug 1 1");
#	warnfprint(ctlfd, "debug 2 1");
	pidc := chan of int;
	spawn reader(pidc);
	readerpid = <- pidc;
	return 0;
}

shutdown()
{
	if (readerpid >= 0)
		kill(readerpid);
}
