implement Styxd;

include "sys.m";
	sys: Sys;

include "draw.m";

include "keyring.m";

include "security.m";

include "arg.m";

stderr: ref Sys->FD;

Styxd: module 
{
	init: fn(ctxt: ref Draw->Context, args: list of string);
};

usage()
{
	sys->fprint(stderr, "usage: styxd [-n] [-r rootdir] [-s] alg ...\n");
	raise "fail:usage";
}

init(nil: ref Draw->Context, args: list of string)
{
	sys = load Sys Sys->PATH;
	stdin := sys->fildes(0);
	stderr = sys->open("/dev/cons", Sys->OWRITE);

	setid := 1;
	rootdir := "/";
	newns := Sys->FORKNS | Sys->FORKENV;

	arg := load Arg Arg->PATH;
	if(arg == nil)
		err(sys->sprint("can't load %s: %r", Arg->PATH));
	arg->init(args);
	while((opt := arg->opt()) != 0)
		case opt {
		's' =>		# temporary, undocumented option
			setid = 0;
		'n' =>
			newns = 0;
		'r' =>
			rootdir = arg->arg();
			if(rootdir == nil)
				usage();
		* =>
			usage();
		}
	args = arg->argv();
	arg = nil;

	if(args == nil)
		err("no algorithm list");

	auth := load Auth Auth->PATH;
	if(auth == nil)
		err(sys->sprint("can't load %s: %r", Auth->PATH));

	error := auth->init();
	if(error != nil)
		err(sys->sprint("Auth init failed: %s", error));

	user := user();

	kr := load Keyring Keyring->PATH;
	ai := kr->readauthinfo("/usr/"+user+"/keyring/default");
	# let auth->server handle nil ai


	(fd, info_or_err) := auth->server(args, ai, stdin, setid);
	if(fd == nil )
		err(sys->sprint("authentication failed: %s", info_or_err));

	if(newns)
		sys->pctl(newns, nil);

	if(sys->export(fd, rootdir, Sys->EXPASYNC) < 0)
		sys->fprint(sys->fildes(2), "styxd: file export: %r\n");
}

user(): string
{
	fd := sys->open("/dev/user", Sys->OREAD);
	if(fd == nil)
		err(sys->sprint("can't open /dev/user: %r"));

	buf := array[Sys->NAMEMAX] of byte;
	n := sys->read(fd, buf, len buf);
	if(n < 0)
		err(sys->sprint("error reading /dev/user: %r"));

	return string buf[0:n];	
}

err(s: string)
{
	sys->fprint(stderr, "styxd: %s\n", s);
	raise "fail:error";
}
