
#include "dat.h"
#include "fns.h"
#include "isa.h"
#include "interp.h"
#include "kernel.h"

void **Xup;

enum
{
	KSTACK = 32*1024,
};

extern void tramp(char*, void(*)(void*), void*);

void
cleanexit(int x)
{
	USED(x);
//	print("clean exit: %d\n", x);
//	killrefresh();
	postnote(PNGROUP, getpid(), "interrupt");
	exits("interrupt");
}

int
kclose(int fd)
{
	return close(fd);
}

Dir*
kdirfstat(int fd)
{
	Dir *dd;

	if((dd = dirfstat(fd)) == nil){
		rerrstr(up->env->errstr, sizeof up->env->errstr);
		return nil;
	}

	dd->name = "dis";
	dd->uid = "dis";
	dd->gid = "dis";
	dd->muid = "dis";
	return dd;
}
	
int
kopen(char *path, int mode)
{
	int fd;

	fd = open(path, mode);
	if(fd < 0)
		rerrstr(up->env->errstr, sizeof up->env->errstr);
	return fd;
}

int
kproc(char *name, void (*func)(void*), void *arg, int flags)
{
	int pid;
	Proc *p;

	USED(flags);

// print("kproc %s\n", name);
	p = newproc();
	p->kstack = mallocz(KSTACK, 0);
	if(p == nil || p->kstack == nil)
		panic("kproc: no memory");

/*
	p->env->uid = up->env->uid;
	p->env->gid = up->env->gid;
	memmove(p->env->user, up->env->user, NAMELEN);
*/

	strcpy(p->text, name);

	p->func = func;
	p->arg = arg;

	lock(&procs.l);
	if(procs.tail != nil) {
		p->prev = procs.tail;
		procs.tail->next = p;
	}
	else {
		procs.head = p;
		p->prev = nil;
	}
	procs.tail = p;
	unlock(&procs.l);

	/*
	 * switch back to the unshared stack to do the fork
	 * only the parent returns from kproc
	 */
	up->kid = p;
	up->kidsp = p->kstack;
	pid = setjmp(up->sharestack);
	if(!pid)
		longjmp(up->privstack, 1);
	return pid;
}

void
krendez(Proc *p, ulong val)
{
	rendezvous((ulong)p, val);
}

long
kread(int fd, void *data, long count)
{
	count = read(fd, data, count);
	if(count < 0)
		rerrstr(up->env->errstr, sizeof up->env->errstr);
	return count;
}

long
kwrite(int fd, void *data, long count)
{
	count = write(fd, data, count);
	if(count < 0)
		rerrstr(up->env->errstr, sizeof up->env->errstr);
	return count;
}

vlong	kseek(int fd, vlong n, int type) {
	return seek(fd, n, type);
}

int kpipe(int fd[2]) {
	return pipe(fd);
}

int kdup(int oldfd, int newfd) {
	return dup(oldfd, newfd);
}

int kcreate(char* file, int omode, ulong perm) {
	return create(file, omode, perm);
}

int kremove(char* file) {
	return remove(file);
}

int kunmount(char* name, char* old) {
	return unmount(name, old);
}

long kpread(int fd, void* buf, long nbytes, vlong offset) {
	return pread(fd, buf, nbytes, offset);
}

long kpwrite(int fd, void* buf, long nbytes, vlong offset) {
	return pwrite(fd, buf, nbytes, offset);
}

int kchdir(char* file) {
	return chdir(file);
}

Dir* kdirstat(char* file) {
	return dirstat(file);
}

char* kfd2path(int fd) {
	static char buf[128];

	fd2path(fd, buf, sizeof buf);

	return buf;
}

int kmount(int fd, int afd, char* old, int flag, char* aname) {
	return mount(fd, afd, old, flag, aname);
}

int kbind(char* name, char* old, int flag) {
	return bind(name, old, flag);
}

int kdirwstat(char* name, Dir* dir) {
	return dirwstat(name, dir);
}

int kdirfwstat(int fd, Dir* dir) {
	return dirfwstat(fd, dir);
}

int kdial(char* addr, char* local, char* dir, int* cfdp) {
	return dial(addr, local, dir, cfdp);
}

int kannounce(char* addr, char* dir) {
	return announce(addr, dir);
}

int klisten(char* dir, char* newdir) {
	return listen(dir, newdir);
}

long kdirread(int fd, Dir** buf) {
	return dirread(fd, buf);
}

int kfauth(int fd, char* aname) {
	return fauth(fd, aname);
}

int   kfversion(int fd, uint bufsize, char *version, uint nversion) {
	return fversion(fd, bufsize, version, nversion);
}

static void
traphandler(void *reg, char *msg)
{
	int intwait = up->intwait;
	up->intwait = 0;
	/* Ignore pipe writes from devcmd */
	if(strstr(msg, "write on closed pipe") != nil)
		noted(NCONT);

/*
	if (sflag) {
		if (intwait && strcmp(msg, Eintr) == 0)
			noted(NCONT);
		else
			noted(NDFLT);
	}
*/
	if(intwait == 0)
		disfault(reg, msg);
	noted(NCONT);
}

void
libinit(char *imod)
{
	char *sp;
	Proc *xup, *p;
	int pid;

	rfork(RFNAMEG|RFREND|RFNOTEG);

//	if(sflag == 0)
		notify(traphandler);

	Xup = &xup;

	/*
	 * dummy up a up and stack so the first proc
	 * calls emuinit after setting up his private jmp_buf
	 */
	p = newproc();
	p->kstack = mallocz(KSTACK, 0);
	if(p == nil || p->kstack == nil)
		panic("libinit: no memory");
	sp = p->kstack;
	p->func = emuinit;
	p->arg = imod;

	/*
	 * set up a stack for forking kids on separate stacks.
	 * longjmp back here from kproc.
	 */
	while(setjmp(p->privstack)){
		p = up->kid;
		sp = up->kidsp;
		switch(pid = rfork(RFPROC|RFMEM)){
		case 0:
			/*
			 * send the kid around the loop to set up his private jmp_buf
			 */
			break;
		default:
			/*
			 * parent just returns to his shared stack in kproc
			 */
			longjmp(up->sharestack, pid);
			panic("longjmp failed");
		}
	}

	/*
	 * you get here only once per Proc
	 * go to the shared memory stack
	 */
	up = p;
	up->sigid = getpid();
//	print("*** %d: %d\n", getpid(), up->pid);
	tramp(sp+KSTACK, up->func, up->arg);
	panic("tramp returned");
}

void
oshostintr(Proc *p)
{
	postnote(PNPROC, p->sigid, "interrupt");
}

void
oslongjmp(void *ureg, osjmpbuf jmp, int val)
{
	if(ureg)
		notejmp(ureg, jmp, val);
	else
		longjmp(jmp, val);
}

void
ospause(void)
{
	for(;;)
		sleep(1000000);
}

void
osyield(void)
{
	sleep(0);
}

int
osmillisleep(ulong milsec)
{
	sleep(milsec);
	return 0;
}

void
osready(Proc *p)
{
//	print("osready:%d:%p\n", getpid(), p);
	rendezvous((ulong)p, 0);
}

void
osblock(void)
{
//	print("osblock:%d:%p\n", getpid(), up);
	rendezvous((ulong)up, 0);
}

/* void sysinit() {
}*/
static vlong
b2v(uchar *p)
{
	int i;
	vlong v;

	v = 0;
	for(i=0; i<sizeof(uvlong); i++)
		v = (v<<8)|p[i];
	return v;
}

long
osmillisec(void)
{
	int n;
	static vlong nsec0 = 0;
	static int nsecfd = -1;
	uchar buf[sizeof(uvlong)];

	if(nsec0 == 0){
		nsecfd = open("/dev/bintime",OREAD);  /* never closed */
		if(nsecfd<0){
			fprint(2,"can't open /dev/bintime: %r\n");
			return(0);
		}
		n = read(nsecfd,buf,sizeof(buf));
		if(n != sizeof(buf)){
			fprint(2,"read err on /dev/bintime: %r\n");
			return(0);
		}
		nsec0 = b2v(buf);
		return(0);
	}
	n = read(nsecfd, buf, sizeof(buf));
	if(n!=sizeof(buf)) {
		fprint(2,"read err on /dev/bintime: %r\n");
		return(0);
	}
	return ((b2v(buf)-nsec0)/1000000);
}

int
limbosleep(ulong milsec)
{
	return osmillisleep(milsec);
}
