#include	"dat.h"
#include	"fns.h"
#include	"error.h"
#include	"interp.h"
#include	"kernel.h"
#include	"version.h"

void
error(char *err)
{
//	print("error:%d:%s\n", getpid(), err);
	kwerrstr(err);
	nexterror();
}

Proc*
newproc(void)
{
	Proc *p;

	p = malloc(sizeof(Proc));
	if(p == nil)
		return nil;

	p->type = Unknown;
	p->killed = 0;
	p->swipend = 0;
	p->env = &p->defenv;
	kstrdup(&p->env->user, "*nouser");
	p->env->errstr = p->env->errbuf0;
	p->env->syserrstr = p->env->errbuf1;
	addprog(p);

//	print("newproc: %p: %d\n", p, getpid());

	return p;
}

void
osenter(void)
{
	up->syscall = 1;
}

void
osleave(void)
{
	int r;

	lock(&up->rlock);
	r = up->swipend;
	up->swipend = 0;
	unlock(&up->rlock);

	lock(&up->sysio);
	up->syscall = 0;
	unlock(&up->sysio);

	/* Cleared by the signal/note/exception handler */
	while(up->intwait)
		osyield();

	if(r != 0)
		error(Eintr);
}

void
Sleep(Rendez *r, int (*f)(void*), void *arg)
{
//	print("sleep:%d: %p\n", getpid(), up);
	lock(&up->rlock);
	lock(&r->l);

	/*
	 * if interrupted or condition happened, never mind
	 */
	if(up->killed || f(arg)) {
		unlock(&r->l);
	}else{
		if(r->p != nil)
			panic("double sleep pc=0x%lux %s[%lud] %s[%lud] r=0x%lux\n", getcallerpc(&r), r->p->text, r->p->pid, up->text, up->pid, r);

		r->p = up;
		unlock(&r->l);
		up->swipend = 0;
		up->r = r;	/* for swiproc */
		unlock(&up->rlock);

		osblock();

		lock(&up->rlock);
		up->r = nil;
	}

	if(up->killed || up->swipend) {
		up->killed = 0;
		up->swipend = 0;
		unlock(&up->rlock);
		error(Eintr);
	}
	unlock(&up->rlock);
}

int
Wakeup(Rendez *r)
{
	Proc *p;

//	print("wakeup:%d: %p\n", getpid(), r->p);
	lock(&r->l);
	p = r->p;
	if(p != nil) {
		r->p = nil;
		osready(p);
	}
	unlock(&r->l);
	return p != nil;
}

void
swiproc(Proc *p, int interp)
{
	Rendez *r;
	
	if(p == nil)
		return;

	/*
	 * Pull out of emu Sleep
	 */
	lock(&p->rlock);
	if(!interp)
		p->killed = 1;
	r = p->r;
	if(r != nil) {
		lock(&r->l);
		if(r->p == p){
			p->swipend = 1;
			r->p = nil;
			osready(p);
		}
		unlock(&r->l);
		unlock(&p->rlock);
		return;
	}
	unlock(&p->rlock);

	/*
	 * Maybe pull out of Host OS
	 */
	lock(&p->sysio);
	if(p->syscall && p->intwait == 0) {
		p->swipend = 1;
		p->intwait = 1;
		unlock(&p->sysio);
		oshostintr(p);
		return;	
	}
	unlock(&p->sysio);
}



void
nexterror(void)
{
	oslongjmp(nil, up->estack[--up->nerr], 1);
}

/* for dynamic modules - functions not macros */

void*
waserr(void)
{
	up->nerr++;
	return up->estack[up->nerr-1];
}

void
poperr(void)
{
	up->nerr--;
}

void
panic(char *fmt, ...)
{
	va_list arg;
	char buf[512];

	va_start(arg, fmt);
	vseprint(buf, buf+sizeof(buf), fmt, arg);
	va_end(arg);
	fprint(2, "panic: %s\n", buf);
//	if(sflag)
		abort();
//	ospause();

	cleanexit(0);
}

/* char*
syserr(char *s, char *es, Prog *p)
{
	Osenv *o;

	o = p->osenv;
	kstrcpy(s, o->errstr, es - s);
	return s + strlen(s);
} */

WORD
modstatus(REG *r, char *ptr, int len)
{
	Inst *PC;
	Frame *f;

	if(r->M->m->name[0] == '$') {
		f = (Frame*)r->FP;
		snprint(ptr, len, "%s[%s]", f->mr->m->name, r->M->m->name);
		if(f->mr->compiled)
			return (WORD)f->lr;
		return f->lr - f->mr->prog;
	}
	memmove(ptr, r->M->m->name, len);
	if(r->M->compiled)
		return (WORD)r->PC;
	PC = r->PC;
	/* should really check for blocked states */
	if(PC > r->M->prog)
		PC--;
	return PC - r->M->prog;
}

/* 
int
incref(Ref *r)
{
	int x;

	lock(&r->lk);
	x = ++r->ref;
	unlock(&r->lk);
	return x;
}

int
decref(Ref *r)
{
	int x;

	lock(&r->lk);
	x = --r->ref;
	unlock(&r->lk);
	if(x < 0) 
		panic("decref, pc=0x%lux", getcallerpc(&r));

	return x;
}
*/

/*
 * Rather than strncpy, which zeros the rest of the buffer, kstrcpy
 * truncates if necessary, always zero terminates, does not zero fill,
 * and puts ... at the end of the string if it's too long.  Usually used to
 * save a string in up->genbuf;
 */
void
kstrcpy(char *s, char *t, int ns)
{
	int nt;

	nt = strlen(t);
	if(nt+1 <= ns){
		memmove(s, t, nt+1);
		return;
	}
	/* too long */
	if(ns < 4){
		/* but very short! */
		strncpy(s, t, ns);
		return;
	}
	/* truncate with ... at character boundary (very rare case) */
	memmove(s, t, ns-4);
	ns -= 4;
	s[ns] = '\0';
	/* look for first byte of UTF-8 sequence by skipping continuation bytes */
	while(ns>0 && (s[--ns]&0xC0)==0x80)
		;
	strcpy(s+ns, "...");
}

/*
 * Atomically replace *p with copy of s
 */
void
kstrdup(char **p, char *s)
{
	int n;
	char *t, *prev;
	static Lock l;

	n = strlen(s)+1;
	t = kmalloc(n);
	if(t == nil)
		panic("kstrdup: no memory");
	memmove(t, s, n);
	prev = *p;
	*p = t;
	free(prev);
}

char*
enverror(void)
{
	return up->env->errstr;
}

#define	QP(l)	(Prog**)((char*)(l)+sizeof(QLock))

void*
libqlowner(void *l)
{
	return *QP(l);
}

void
libqlock(void *l)
{
	Prog *p;
	QLock *q;

	q = l;
	p = currun();
	if(p == nil)
		abort();

	if(!canqlock(q)) {
		release();
		qlock(q);
		acquire();
	}
	*QP(l) = p;
}

void
libqunlock(void *l)
{
	Prog *p;
	QLock *q;

	q = l;
	p = currun();
	if(p == nil)
		abort();
	if(*QP(l) != p)
		abort();

	*QP(l) = nil;
	qunlock(q);
}

void*
libqlalloc(void)
{
	QLock *q;

	q = mallocz(sizeof(QLock)+sizeof(Prog*), 1);
	return q;
}

void
libqlfree(void *l)
{
	free(l);
}

void
rptwakeup(void *o, Rept *r)
{
	if(r == nil)
		return;
	lock(&r->l);
	r->o = o;
	unlock(&r->l);
	Wakeup(&r->r);
}

static int
rptactive(void *a)
{
	Rept *r = a;
	int i;
	lock(&r->l);
	i = r->active(r->o);
	unlock(&r->l);
	return i;
}

static void
rproc(void *a)
{
	long now, then;
	ulong t;
	int i;
	void *o;
	Rept *r;

//	up->env->fgrp = newfgrp(nil);
	r = a;
	t = (ulong)r->t;

Wait:
	Sleep(&r->r, rptactive, r);
	lock(&r->l);
	o = r->o;
	unlock(&r->l);
	then = osmillisec();
	for(;;){
		osmillisleep(t);
		now = osmillisec();
		if(waserror())
			break;
		i = r->ck(o, now-then);
		poperror();
		if(i == -1)
			goto Wait;
		if(i == 0)
			continue;
		then = now;
		acquire();
		if(waserror()) {
			release();
			break;
		}
		r->f(o);
		poperror();
		release();
	}
//	pexit("", 0);
}

Rept*
rptproc(char *s, int t, void *o, int (*active)(void*), int (*ck)(void*, int), void (*f)(void*))
{
	Rept *r;

	r = mallocz(sizeof(Rept), 1);
	if(r == nil)
		return nil;
	r->t = t;
	r->active = active;
	r->ck = ck;
	r->f = f;
	r->o = o;
	kproc(s, rproc, r, KPDUPPG);
	return r;
}
